#! /usr/bin/env python
import sys
import re
from Bio import SeqIO
from Bio.Blast import NCBIXML

###################################  CONTACT, QUESTIONS, COMMENTS, ADVICE:  ##########################################

#Ben Ewen-Campen = bewencampen@oeb.harvard.edu

#################################### DESCRIPTION: ##########################################
#This program takes a set of one or more .fasta files and their associated BLAST results (in XML format)
#and creates several annotated output files which are more easily read by humans:

#################################### OUTPUT FILES: ##########################################
#Output file 1: suffix='_annotated.fasta' 
	# The input sequences (e.g. isotigs/contigs/singletons) are renamed to include top 20 BLAST hits, or else "*NO HIT*"

#Output file 2: suffix='_blast_statistics.txt'= 
	#select summary statistics, including percentage of sequences with BLAST hits, number of repeat hits to the same BLAST hits, and number of possible paralogs.
		
#Output file 3: suffix='_blast.tabdelimited'
	#A tab-delimited file containing information about the top BLAST hit for each sequence.  
	#The columns are: 'Sequence ID [e.g. Isotig0001]', 'Sequence Length', 'Has BLAST Hit? [Y/N]', 'Top BLAST Hit', 'E-value', 'Score', 'Subject Start Position', 'Subject End Position'

#Output file 4: suffux='_unique_seqs_list.txt'
	#a list of sequence names that excludes any having a previously-seen top BLAST hit.  This list can be used in the program 'blast2GO' to select only one sequence per top BLAST hit.

###################################  DEFINITIONS:  ##########################################
#DEFINITIONS OF 'REPEAT','PARALOG',and 'IDENTICAL'
	#'REPEAT' = If two sequences have the same top BLAST hit and they hit non-overlapping portions of that top BLAST hit, they are called 'REPEAT'.  This implies that they represent unassembled portions of the same transcript and should not be double-counted during summarizing results.
	#'PARALOG' = If two sequences have the same top BLAST hit and they hit *overlapping* portions of that BLAST hit, they are called 'PARALOGS'.  This implies that the two sequences represent different transcripts, and therefore should each be counted in summary statistics.
	#'IDENTICAL' = If two sequences have the same top BLAST hit and they hit the *exact* same portion of that BLAST hit, they are counted as PARALOGS in this script.  This category is only included to help diagnose if you have many identical sequences in your transcriptome.  For instance, the singletons generated by Newbler v2.3 often include several identical sequences which may be identified using this category.

###################################  IMPORTANT NOTES ABOUT RUNNING: #########################################

#NOTE: Currently, this program requires .fasta file and blastresult file to be in identical orders.
#NOTE: Both the .fasta file(s) and the .xml file(s) must be named sequentially i.e. 'seqs_1.fasta', 'seqs_2.fasta', 'seqs_3.fasta'
#NOTE: Currently, the program requires the input sequences to be named in a specific way so that they may be properly summarized in the 'blast_statistics.txt' file:
		#Isotigs and contigs must have names that begin with "isotig" or "contig"
		#Cap3-assembled singletons must have names that begin with 'cap3_' i.e. 'cap3_contigs' and/or 'cap3_singletons' would both work.


usage="""
transcriptome_blast_summarizer.py

Usage: 

transcriptome_blast_summarizer.py yourcontigfasta yourcontigblast yoursingletonfasta yoursingletonblast number_of_contig_files[interger] number_of_singletons_files [interer] outfile
	
"""

#remind user of usage

if len(sys.argv)<7:
	print usage	
	
#the input files

else:
	
	contigfastafile = sys.argv[1]
	contigblastfile = sys.argv[2]
	singletonfastafile = sys.argv[3]
	singletonblastfile = sys.argv[4]
	numcontigfiles = int(sys.argv[5])
	numsingletonfiles = int(sys.argv[6])
	outfile = sys.argv[7]
	
	fastainsuffix=".fasta"
	blastsuffix=".xml"
	
	#suffixes to add to the output files:
	fastasuffix='_annotated.fasta'
	summarysuffix='_blast_statistics.txt'
	tabsuffix='_tab_delimited.txt'
	uniquesequencelistsuffix='_unique_seq_list.txt'
	repeatsuffix='_repeats_summary.txt'

	
	fastaouthandle = open(outfile + fastasuffix, "w")
	tabouthandle=open(outfile + tabsuffix, "w")
	summaryouthandle=open(outfile + summarysuffix, "w")
	uniquesequencelisthandle=open(outfile + uniquesequencelistsuffix, "w")
	repeatsummaryhandle=open(outfile + repeatsuffix, "w")

	
	
	#which columns to put in the tab-delimited output? 
	tabcolumns=['ID', 'Length', 'Has BLAST Hit?', 'Top Hit', 'E-value', 'Score', 'Subject Start', 'Subject End']
	tabheader="\t".join(tabcolumns)
	tabouthandle.write(tabheader+'\n')
	
	# initialize the summary statistics
	numseqs=0 
	numsingletons=0
	numseqs_with_hits=0
	numseqs_without_hits=0
	num_overlapping=0
	unique_sequences=0
	repeat_hits=0
	num_identical=0
	
	numcontigs=0
	numcontigs_with_hits=0
	numcontigs_repeats=0
	numcontigs_overlapping=0
	numcontigs_unique=0
	
	numsingletons=0
	numsingletons_with_hits=0
	numsingletons_repeats=0
	numsingletons_overlapping=0
	numsingletons_unique=0
	

	#create a dictionary where keys are blast hits and values are contigs, which we then use to figure out unique blast hits

	blastdict={}
	
	#First loop: the isotigs/contigs

	for eachfile in range(numcontigfiles+1):
	
		if eachfile == 0:
		
			continue
			
		else:

			print 'CONTIG FILE ',eachfile
	
			seqs = {} #A dictionary of sequence records keyed by id
			ids = [] #The ids of the sequence in the order they appear in the file, so they can be written back in order
			
			fhandle=open(contigfastafile+str(eachfile)+fastainsuffix, 'r')
			for record in SeqIO.parse(fhandle, 'fasta'):
				seqs[record.id] = record
				ids.append(record.id)
			fhandle.close()
			
			
			#parse the BLAST result
			blast_handle=open(contigblastfile + str(eachfile) + blastsuffix)
			blast_records=NCBIXML.parse(blast_handle)

			# Write the annotated .fasta file and the tab file simultaneously
			# iterate through the input seqs, ask if they have a blast hit, and outwrite top 20 hits or ***NO HIT*** to the sequence desc.
	
			for id in ids:
	
				try:
					blast_record=blast_records.next()
					blast_database=blast_record.database
					blast_evalue_cutoff=blast_record.expect
				except StopIteration:
					break
				
				#count each key in the .fasta input as you iterate though	
				numseqs=numseqs+1
				numcontigs=numcontigs+1

				#put the seqrecord (the value for each key) in a variable called record
				record=seqs[id] 
		
				#initialize the list of hits to write to the annotated .fasta
				top5hits=[]
		
				#a variable to count through the five hits to label the output
				n=1
		
				#write the .fasta and .tab outputs: 
				#if there is no blast hit, it will raise an IndexError, so we use a try/except thingy:		
				
				try:
					blast_record.descriptions[0]!=''
					for description in blast_record.descriptions[0:20]: #if you want fewer than 20 BLAST hits written, change this value.
						hitcounter='**HIT %s** ' % (n)
						top5hits.append(str(hitcounter))
						top5hits.append(str(description))
						top5hits.append(' ')
						n=n+1
						
					top5string=''.join(top5hits)
				
					fastaouthandle.write('>' + record.description + '\tTop 20 BLAST hits: ' + top5string + '\n' + str(record.seq) + '\n')
				
					for alignment in blast_record.alignments[0:1]:
						blasthit=alignment.title
						for hsp in alignment.hsps[0:1]:
							topevalue=str(hsp.expect)
							topscore=str(hsp.score)
							topstart=hsp.sbjct_start
							topstop=hsp.sbjct_end
						for hsp in alignment.hsps[1:]: #check to see the subsequent hsps have extra length (different hsps arise, eg with frameshifts)
							topstart=min(topstart, hsp.sbjct_start)
							topstop=max(topstop, hsp.sbjct_end)
											
											
					tabouthandle.write(record.name+'\t'+str(len(record.seq))+'\tY\t'+blasthit+'\t'+str(topevalue)+'\t'+str(topscore)+'\t'+str(topstart)+'\t'+str(topstop)+'\n')
						
					try:
						#Test if blasthit already exists. If hit is unique, it raises a KeyError and skips this loop.
							
							
						blastdict[str(blasthit)].append((str(record.id), topstart, topstop))					

						
					except KeyError:


						blastdict[str(blasthit)]=[(str(record.id), topstart, topstop)]

					numseqs_with_hits=numseqs_with_hits+1	
					numcontigs_with_hits=numcontigs_with_hits+1
				
				
				#Here's what will run if sequence has no BLAST hit
				
				except IndexError:
			
					fastaouthandle.write('>'+record.description+'\t***No hits to '+blast_database+ ' at e-value cutoff of: ' +blast_evalue_cutoff+ '***'+'\n'+str(record.seq)+'\n')
		
					tabouthandle.write(record.name+'\t'+str(len(record.seq))+'\tN'+'\n')
			
					numseqs_without_hits = numseqs_without_hits+1
					
					continue


	#ROUND 2, THROUGH THE SINGLETONS:
	
	for eachfile in range(numsingletonfiles+1):
	
		if eachfile == 0:
		
			continue
			
		else:
			print 'SINGLETON FILE ',eachfile #include this so that if a blast result is empty, you'll know which one
	
			seqs = {} #A dictionary of sequence records keyed by id
			ids = [] #The ids of the sequence in the order they appear in the file, so they can be written back in order
			
			fhandle=open(singletonfastafile+str(eachfile)+fastainsuffix, 'r')
			for record in SeqIO.parse(fhandle, 'fasta'):
				seqs[record.id] = record
				ids.append(record.id)
			fhandle.close()
			
			
			#parse the BLAST result
			blast_handle=open(singletonblastfile + str(eachfile) + blastsuffix)
			blast_records=NCBIXML.parse(blast_handle)

			# Write the annotated .fasta file and the tab file simultaneously
			# iterate through the input seqs, ask if they have a blast hit, and outwrite top 5 hits or ***NO HIT*** to the sequence desc.
	
			# The fasta_index is a dictionary, so you need to go through each 'key' and assign the value to a new variable 'record'
			# then treat the 'record' as a callable Seq object.

			for id in ids:
	
				try:
					blast_record=blast_records.next()
					blast_database=blast_record.database
					blast_evalue_cutoff=blast_record.expect
				except StopIteration:
					break
				
				#count each key in the .fasta input as you iterate though	
				numseqs=numseqs+1
				numsingletons=numsingletons+1
				#put the seqrecord (the value for each key) in a variable called record
				record=seqs[id] 
		
				#initialize the list of hits to write to the annotated .fasta
				top5hits=[]
		
				#a variable to count through the five hits to label the output
				n=1
		
				#write the .fasta and .tab outputs: 
				#if there is no blast hit, it will raise an IndexError, so we use a try/except thingy:		
				
				try:
					blast_record.descriptions[0]!=''
					for description in blast_record.descriptions[0:20]:
						hitcounter='**HIT %s** ' % (n)
						top5hits.append(str(hitcounter))
						top5hits.append(str(description))
						top5hits.append(' ')
						n=n+1
						
					top5string=''.join(top5hits)
				
					fastaouthandle.write('>' + record.description + '\tTop 20 BLAST hits: ' + top5string + '\n' + str(record.seq) + '\n')
				
					#tabcolumns=['ID', 'Length', 'Has BLAST Hit?', 'Top Hit', 'E-value', 'Score', 'Subject Start', 'Subject End']
					for alignment in blast_record.alignments[0:1]:
						blasthit=alignment.title
						for hsp in alignment.hsps[0:1]:
							topevalue=str(hsp.expect)
							topscore=str(hsp.score)
							topstart=hsp.sbjct_start
							topstop=hsp.sbjct_end
						for hsp in alignment.hsps[1:]: #to account for frameshifts etc, take min and max subct_start and stop for all hsps
							topstart=min(topstart, hsp.sbjct_start)
							topstop=max(topstop, hsp.sbjct_end)
									
					tabouthandle.write(record.name+'\t'+str(len(record.seq))+'\tY\t'+blasthit+'\t'+str(topevalue)+'\t'+str(topscore)+'\t'+str(topstart)+'\t'+str(topstop)+'\n')
						
					try:
						#Test if blasthit already exists. If hit is unique, it raises a KeyError and skips this loop.
														
						blastdict[str(blasthit)].append((str(record.id), topstart, topstop))					

						
					except KeyError:


						blastdict[str(blasthit)]=[(str(record.id), topstart, topstop)]

					numseqs_with_hits=numseqs_with_hits+1	
					numsingletons_with_hits=numsingletons_with_hits+1
				
				
				#Here's what will run if sequence has no BLAST hit
				
				except IndexError:
			
					fastaouthandle.write('>'+record.description+'\t***No hits to '+blast_database+ ' at e-value cutoff of: ' +blast_evalue_cutoff+ '***'+'\n'+str(record.seq)+'\n')
		
					tabouthandle.write(record.name+'\t'+str(len(record.seq))+'\tN'+'\n')
			
					numseqs_without_hits = numseqs_without_hits+1
					
					continue
	
	
	
	#NOW DEAL WITH THE UNIQUES,REPEATS(including identicals),PARALOGS,
	
	for key in blastdict.keys():
	
		#first write one unique hit her blasthit
		
		uniquesequencelisthandle.write(blastdict[key][0][0]+'\n')
		repeatsummaryhandle.write('UNIQUE\t'+blastdict[key][0][0]+'\n')

		unique_sequences=unique_sequences+1
#		print 'adding round1 of unique hits: ', unique_sequences
		#if there is only one sequence per blasthit, count it as unique
		
#		if len(blastdict[key]) == 1:
		
#			continue
#			print key, blastdict[key]
		
		#if there are more than one sequence, figure out if it overlaps (paralog) or not (repeat)
		
		if len(blastdict[key]) > 1:
		
			#make a growing list of starts and stop instead of single values!
			
			print key, blastdict[key]
			startstop=[(blastdict[key][0][1], blastdict[key][0][2])]
			
			#iterate through the values of a key e.g. 0=(contig1, 3, 85),1=(contig2, 90, 150) etc.
			
			for repeat in range(len(blastdict[key])):
				
				#we initiated the startstop with the start and stop of the first blast hit, so skip it here
				if repeat == 0:
					continue
				
				#now test each subsequent value against all the existing startstops 
				else:
					for i in range(len(startstop)):
						if startstop[i][0] == blastdict[key][repeat][1] and startstop[i][1] == blastdict[key][repeat][2]:
							print 'IDENTICAL HIT: ', blastdict[key][repeat][0], 'hits ', key, 'from ', blastdict[key][repeat][1], '==>', blastdict[key][repeat][2], '.  It was already hit from', startstop[i][0],  '==>', startstop[i][1] 
							repeatsummaryhandle.write('IDENTICAL\t'+blastdict[key][repeat][0]+'\t'+key+'\n')	
							#if i == 0:
								
							num_identical=num_identical+1
							unique_sequences=unique_sequences+1
#								print 'adding identical uniques: ', unique_sequences
							
							uniquesequencelisthandle.write(blastdict[key][repeat][0]+'\n')
							num_overlapping=num_overlapping+1
								
							break #stop scanning startstops, move to next repeat
						
						if startstop[i][0] <= blastdict[key][repeat][1] <= startstop[i][1] or startstop[i][0] <= blastdict[key][repeat][2] <= startstop[i][1] or (blastdict[key][repeat][1] <= startstop[i][0] and blastdict[key][repeat][2] >= startstop[i][1]):
							print 'POSSIBLE PARALOG: ', blastdict[key][repeat][0], 'hits ', key, 'from ', blastdict[key][repeat][1], '==>', blastdict[key][repeat][2], '.  It was already hit from', startstop[i][0],  '==>', startstop[i][1]
							repeatsummaryhandle.write('PARALOG\t'+blastdict[key][repeat][0]+'\t'+key+'\n')
							startstop[i]=( min(startstop[i][0], blastdict[key][repeat][1]), max(startstop[i][1], blastdict[key][repeat][2]))
							
							#if i == 0:
							unique_sequences=unique_sequences+1
#								print 'adding paralog uniques: ', unique_sequences
							num_overlapping=num_overlapping+1
							
							uniquesequencelisthandle.write(blastdict[key][repeat][0]+'\n')
							
							break #stop scanning startstops, move to next repeat
						
				
						
						else: 
						
							#if blastdict[repeat] doesn't start or stop within an existing hit, append startstop
							#to avoid printing repeat hit multiple times, only print repeat hit when you're done scanning the whole startstop list.
							
#							print 'GOT HERE'
							if i < max(range(len(startstop))):
#								print i, max(range(len(startstop)))
#								print 'ALSO GOT HERE'
								continue
							if i == max(range(len(startstop))): 
#								print 'GOT FINAL'
								startstop.append((blastdict[key][repeat][1],blastdict[key][repeat][2]))
								print 'REPEAT HIT: ', blastdict[key][repeat][0], 'hits ', key, 'from ', blastdict[key][repeat][1], '==>', blastdict[key][repeat][2],'This doesn\'t overlap any previous hits ', startstop
								repeatsummaryhandle.write('REPEAT\t'+blastdict[key][repeat][0]+'\t'+key+'\n')
								repeat_hits=repeat_hits+1
								break
							break


	
	fastaouthandle.close()
	tabouthandle.close()
	uniquesequencelisthandle.close()	
	repeatsummaryhandle.close()
	#get the contig/cap3contigs/cap3singletons stats
	
	read_status=open(outfile + repeatsuffix).read()

	iden_regexs=['IDENTICAL\tisotig|IDENTICAL\tcontig','IDENTICAL\tcap3']
	unique_regexs=['UNIQUE\tisotig|UNIQUE\tcontig','UNIQUE\tcap3']
	paralog_regexs=['PARALOG\tisotig|PARALOG\tcontig','PARALOG\tcap3']
	repeat_regexs=['REPEAT\tisotig|REPEAT\tcontig', 'REPEAT\tcap3']

	num_iden_isotigs=len(re.findall(iden_regexs[0],read_status))
	num_iden_cap3seqs=len(re.findall(iden_regexs[1],read_status))
	
	num_unique_isotigs=len(re.findall(unique_regexs[0], read_status))
	num_unique_cap3seqs=len(re.findall(unique_regexs[1], read_status))
	
	num_paralog_isotigs=len(re.findall(paralog_regexs[0], read_status))
	num_paralog_cap3seqs=len(re.findall(paralog_regexs[1], read_status))
	
	num_repeat_isotigs=len(re.findall(repeat_regexs[0], read_status))
	num_repeat_cap3seqs=len(re.findall(repeat_regexs[1], read_status))

	
	summaryouthandle=open(outfile + summarysuffix, "w")
	
	percentage=float(numseqs_with_hits)/float(numseqs)*100
	
	summaryouthandle.write('BLAST statistics for ' + contigfastafile +' and ' + singletonfastafile) 
	
	summaryouthandle.write('\nDatabase: ' + blast_database + '\nE-value cutoff: ' + blast_evalue_cutoff)
	
	summaryouthandle.write('\n\nALL SEQUENCES (Isotigs+Singletons) STATISTICS:')
	
	summaryouthandle.write('\nTotal number of sequences: ' + str(numseqs) + '\nNumber of sequences with BLAST hits: ' + str(numseqs_with_hits)+ '\nNumber of sequences without BLAST hits: ' + str(numseqs_without_hits) + '\nPercentage of sequences with BLAST hits: ' + str(percentage))
	summaryouthandle.write('\nTotal number of sequences with already-seen top BLAST hits (No. of repeats): ' + str(repeat_hits))	
	summaryouthandle.write('\nTotal number unique BLAST hits (this number includes overlapping hits, ie possible paralogs): ' + str(unique_sequences))
	summaryouthandle.write('\nNumber of sequences that are have overlapping top BLAST hits with another sequence (No. possible paralogs): ' + str(num_overlapping))
	summaryouthandle.write('\nNumber of sequences that have identical hits over the entire hit length (a subset of paralogs which you should look at in depth): ' +str(num_identical))
	
	summaryouthandle.write('\n\nISOTIGS STATISTICS:')
	
	isotighitpercent=float(numcontigs_with_hits)/float(numcontigs)*100
	summaryouthandle.write('\nNumber of isotigs: ' + str(numcontigs) + '\nTotal number of contigs with BLAST hits: ' + str(numcontigs_with_hits))
	summaryouthandle.write('\nPercentage of Isotigs with BLAST hits: ' + str(isotighitpercent))
	summaryouthandle.write('\nNumber of Isotigs with already-seen top BLAST hits (#repeats): ' + str(num_repeat_isotigs)) 
	summaryouthandle.write('\nNumber of Isotigs with overlapping top BLAST hits (paralogs+identicals): ' + str(num_paralog_isotigs+num_iden_isotigs))
	summaryouthandle.write('\nNumber of Isotigs with Unique BLAST hits (uniques+paralogs+identicals): ' + str(num_unique_isotigs+num_paralog_isotigs+num_iden_isotigs))
	summaryouthandle.write('\nNumber of Isotigs that have identical hits over the entire hit length(identicals): ' + str(num_iden_isotigs))
	
	summaryouthandle.write('\n\nCAP3-ASSEMBLED SINGLETONS STATISTICS:')
	#summaryouthandle.write('\nNOTE: Overlapping singletons are counted as REPEATS, not PARALOGS, because I was seeing lots of identical singletons getting counted as unique hits).')
	singletonhitpercent=float(numsingletons_with_hits)/float(numsingletons)*100
	summaryouthandle.write('\nNumber of cap3-assembled singletons: ' + str(numsingletons) + '\nTotal number of singletons with BLAST hits: ' + str(numsingletons_with_hits))
	summaryouthandle.write('\nPercentage of cap3-assembled singletons with BLAST hits: ' + str(singletonhitpercent))
	summaryouthandle.write('\nNumber of singletons with already-seen top BLAST hits (#repeats): ' + str(num_repeat_cap3seqs))
	summaryouthandle.write('\nNumber of singletons with overlapping top BLAST hits (paralogs+identicals): ' + str(num_paralog_cap3seqs+num_iden_cap3seqs))
	summaryouthandle.write('\nNumber of singletons with *Unique* BLAST hits: ' + str(num_unique_cap3seqs+num_paralog_cap3seqs+num_iden_cap3seqs))
	summaryouthandle.write('\nNumber of singletons that have identical hits over the entire hit length: ' + str(num_iden_cap3seqs))
	
	summaryouthandle.close()
	
	print 'DONE.  There should be five output files: ', fastasuffix, summarysuffix, tabsuffix, uniquesequencelistsuffix, repeatsuffix
	print '\n\nTotal # repeats: ', repeat_hits
	print 'Total # unique hits: ', unique_sequences
	print 'Total # hits: ', numseqs_with_hits
	print 'Total # sequences: ', numseqs
	print 'Total # sequences with no hits: ', numseqs_without_hits
	print 'Total # paralogs (overlapping): ', numcontigs_overlapping

